-----------------------------------------------------------------------------
-- |
-- Module      :  LRSX.Interface.DocStrings
-- Copyright   :  (c) D. Sabel, Goethe-University, Frankfurt a.M., Germany
-- License     :  BSD-style 
-- 
-- Maintainer  :  sabel <at> ki.cs.uni-frankfurt.de
--
--  
-- several strings for the command line interface (helptexts)
--
-----------------------------------------------------------------------------
module LRSX.Interface.DocStrings where

-- | the global help text
globalHelp :: String
globalHelp =
 unlines 
    ["Command line interface to the LRSX Reasoning Tool"
    ,"See http://www.ki.cs.uni-frankfurt.de/research/lrsx for more information."
    ,""
    ,"Usage: lrsx COMMAND [FLAGS] [INPUTFILE]"
    ,""
    ,"Commands"
    ,""
    ,"  unify          Unify the equations in INPUTFILE."
    ,"  "
    ,"  match          Match the equations in INPUTFILE."    
    ,"  "
    ,"  overlap        Compute unifiers and critical pairs "
    ,"                 for the rules and overlap-commands in INPUTFILE."
    ,"                 "
    ,"  join           Compute diagrams for the rules and" 
    ,"                 overlap-commands in INPUTFILE."
    ,""
    ,"  convert        Convert between different diagram formats"
    ,"                 (which are computed with the the 'join' command)"
    ,""
    ,"  induct         Use the computed diagrams (see 'lrsx join') and"
    ,"                 run termination provers and certifiers to prove"
    ,"                 that diagram application terminates."
    ,""
    ,"  interactive    Start the interactive mode of the LRSX tool."
    ,""
    ,"  help           Help about commands."
    ,""
    ,"Flags "
    ," "
    ,"  are command specific, type"
    ,"  "
    ,"    lrsx COMMAND help "
    ,"  "
    ,"  for an overview of available flags."
    ,"  "
    ,"(c) David Sabel, 2016"
  ]
  
-- | help text for specific commands
commandHelp :: String -> String


commandHelp "interactive"  =
    unlines 
    ["Start the interactive mode of the LRSX tool"
    ,""
    ,"Usage: lrsx interactive"
    ,""    
    ,"type 'help' in in the interactive mode to display"
    ,"detailed informations for the interactive mode" ]

-- help for the 'overlap' command

commandHelp "overlap"  =
    unlines 
    ["Compute unifiers and critical pairs "
    ,"for the rules and overlap-commands in INPUTFILE."
    ,""
    ,"Usage: lrsx overlap [OPTIONS] INPUTFILE"
    ,""    
    ," OPTIONS"
    ,"" 
    ,"   full-context-guess"
    ,"      guess all context variables to be empty or non-empty"
    ,"      this may increase the number of solutions"
    ,"      (default: off)"
    ,"   "
    ,"   full-environment-guess"
    ,"      guess all enviroment variables to be empty or non-empty"
    ,"      this may increase the number of solutions"
    ,"      (default: off)"
    ,""
    ,"type "
    ,"    lrsx help "
    ,"    "
    ,"  for global help on lrsx"
    ,"  "
    ,"(c) David Sabel, 2016"
  ]

-- help for the 'unify' command
  
commandHelp "unify"  =
 unlines 
    ["Unify equations in INPUTFILE"
    ,""
    ,"Usage: lrsx unify [OPTIONS] INPUTFILE"
    ,""
    ," OPTIONS"
    ,"" 
    ,"   full-context-guess"
    ,"      guess all context variables to be empty or non-empty"
    ,"      this may increase the number of solutions"
    ,"      (default: off)"
    ,""
    ,"   full-environment-guess"
    ,"      guess all enviroment variables to be empty or non-empty"
    ,"      this may increase the number of solutions"
    ,"      (default: off)"
    ,""
    ,"   debug"
    ,"      displays the full unification tableau"
    ,""
    ,"type "
    ,"    lrsx help "
    ,"    "
    ,"  for global help on lrsx"
    ,"  "
    ,"(c) David Sabel, 2016"
  ]
  
commandHelp "match"  =
 unlines 
    ["Match equations in INPUTFILE"
    ,""
    ,"Usage: lrsx match [OPTIONS] INPUTFILE"
    ,""
    ," OPTIONS"
    ,"" 
    ,"   debug"
    ,"      displays the full unification tableau"
    ,""
    ,"type "
    ,"    lrsx help "
    ,"    "
    ,"  for global help on lrsx"
    ,"  "
    ,"(c) David Sabel, 2016"
  ]
  
commandHelp "join"  =
 unlines 
    ["Compute diagrams for the rules and overlap-commands in INPUTFILE."
    ,""
    ,"Usage: lrsx join [FLAGS] INPUTFILE"
    ,""
    ,"FLAGS"
    ,""
    ,"  Output options: (select at most one)"
    ,"     (default: long)"
    ,""
    ,"      long       print all found diagrams separately." 
    ,"                 Output includes also alpha-Rename steps in the diagrams"
    ,""
    ,"      proofs     print diagrams together with detailed proofs"
    ,""
    ,"  Note that shorter representations of the diagrams can be obtained"
    ,"  by running 'lrsx convert' as a second step."
    ,""
    ,"   Additional options:"
    ,""
    ,"      bound=N    set a search-bound for closing the diagram"
    ,"                 smaller bounds will make the search faster, but"
    ,"                 may lead to non-closable cases"
    ,"                 (default=10)"
    ,""
    ,"      expansion-bound=N"
    ,"                 set a bound how often context should be expanded"
    ,"                 according to their grammar."
    ,"                 (default=0)"
    ,""
    ,"      extend-standard-reduction"
    ,"                 allows to add standard reduction to the critical pair"
    ,"                 is only sound, if standard reduction is deterministic"
    ,"                 (default=off, but turned on automatically if the search"
    ,"                               for joining the pair fails)"
    ,"      no-extend-standard-reduction"
    ,"                  forbid to extend the critical pair by standard reductions"
    ,"                  in any case. Adequate for non-deterministic calculi."
    ,"                  (default=off)"
    ,""
    ,"      always-alpha-rename"
    ,"                 for closing diagrams always perform alpha renaming"
    ,"                 (default=off (alpha-renaming is used, but not exclusively)"
    ,""
    ,"      no-alpha-rename"
    ,"                 for closing diagrams never perform alpha renaming"
    ,"                 (default=off)"
    ,""
    ,"      no-stop-on-error"
    ,"                 for closing diagrams do not stop on non-closable cases"
    ,"                 (default=off)"
    ,""
    ,"      full-context-guess"
    ,"                  guess all context variables to be empty or non-empty"
    ,"                  this may increase the number of solutions"
    ,"                  (default: off, but turned on automatically if the search"
    ,"                                 for joining the pair fails)" 
    ,"   "
    ,"      full-environment-guess"
    ,"                  guess all enviroment variables to be empty or non-empty"
    ,"                  this may increase the number of solutions"
    ,"                  (default: off, but turned on automatically if the search"
    ,"                                 for joining the pair fails)" 
    ,""
    ,"Note that the diagrams are printed on stdout."
    ,"Debug information is printed on stderr"
    ,""
    ,"type "
    ,"    lrsx help "
    ,"    "
    ,"  for global help on lrsx"
    ,"  "
    ,"(c) David Sabel, 2016"
  ]


commandHelp "convert"  =
 unlines 
    ["Convert the diagrams in INPUTFILE in another format" 
    ,"and print the converted diagrams to stdout."
    ,""
    ,"Usage: lrsx convert FROM TO [OPTIONS] INPUTFILE"
    ,""
    ,"Formats for FROM and TO"
    ,""
    ,"      short      a short representation of the diagrams"
    ,""
    ,"      trs        print diagrams as a term rewrite system"
    ,"                 (with free variables on right hand sides"
    ,"                  to guess transitive closures)"
    ,""
    ,"      itrs       print diagrams as a integer term rewrite system"
    ,"                 (with free integer variables on right hand sides"
    ,"                  to guess transitive closures)"
    ,""
    ,"      long       print all found diagrams separately, includes "
    ,"                 also alpha-Rename steps in the diagrams"
    ,""
    ,"      proofs     print diagrams together with detailed proofs"
    ,""
    ,"Available conversions:"
    ,""
    ,"     FROM   | TO   "
    ,"     -------+------"
    ,"     proofs | long "
    ,"     proofs | short"
    ,"     proofs | trs"
    ,"     proofs | itrs"
    ,"     long   | short"
    ,"     long   | trs"
    ,"     long   | itrs"
    ,"     short  | trs"
    ,"     short  | itrs"
    ,""
    ," OPTIONS"
    ,"" 
    ,"  unions-from=FILE"
    ,"      Use the union and expand commands found in file FILE" 
    ,"      for unifying reduction and transformation rules and "
    ,"      add rules for the expansions"
    ,"      Note that this option is ignored for conversions from"
    ,"      proofs to long!"
    ,""

    ,"type "
    ,"    lrsx help "
    ,"    "
    ,"  for global help on lrsx"
    ,"  "
    ,"(c) David Sabel, 2016"
  ]    
  
commandHelp "induct"  =
 unlines 
    ["Perform the automated induction for a set of diagrams."
    ,"Note that this command will produce several output files."
    ,""
    ,"Usage: lrsx induct [INPUTFORMAT] [OPTIONS] INPUTFILE"
    ,""
    ,"The diagrams from INPUTFILE will be read, transformed in"
    ,"term rewrite systems, and then termination will be proved"
    ,"and certified by external tools"    
    ,""
    ,"INPUTFORMAT"
    ,"      (default=Long)"
    ,""
    ,"      short      a short representation of the diagrams"
    ,""
    ,"      long       print all found diagrams separately, includes "
    ,"                 also alpha-Rename steps in the diagrams"
    ,""
    ,"      proofs     print diagrams together with detailed proofs"
    ,""
    ,"OPTIONS"
    ,""
    ,"      atp-path=PATH"
    ,"                 specify the path to AProVE and CeTA"
    ,"      unions-from=FILE"
    ,"                 Use the union and expand commands found in file FILE" 
    ,"                 for unifying reduction and transformation rules and "
    ,"                 add rules for the expansions"
    ,"      output-to=PATH"
    ,"                 specify the path where the output will be written (default = .)"
    ,"      use-itrs"
    ,"                 use integer term rewrite systems as target"
    ,"                 (note that the certifier CeTA is not run, since it is not able"
    ,"                  to certify such proofs)"
    ,""
    ,"type "
    ,"    lrsx help "
    ,"    "
    ,"  for global help on lrsx"
    ,"  "
    ,"(c) David Sabel, 2016"
  ]    
  
commandHelp _  =
 unlines 
    ["Command not yet implemented"
    ,""
    ,"type "
    ,"    lrsx help "
    ,"    "
    ,"  for global help on lrsx"
    ,"  "
    ,"(c) David Sabel, 2016"
  ]

  
  
  
