> {
> -----------------------------------------------------------------------------
> -- |
> -- Module      :  LRSX.Interface.Parser
> -- Copyright   :  (c) D. Sabel, Goethe-University, Frankfurt a.M., Germany
> -- License     :  BSD-style 
> -- 
> -- Maintainer  :  sabel <at> ki.cs.uni-frankfurt.de
> --
> --  
> -- The Parser of LRSX Unification tool, generated by Happy mainly
> --
> -----------------------------------------------------------------------------
> -- 
> --
> -- other todos:  - add strictness and arity info automatically 
> --               - nub funinfo
> --               - check that all binding positions are correct
> --               - check that strictness information and arities are consistently used
> --               - check that no binding position is strict
> --               - compute the bindarity
> --
> module LRSX.Interface.Parser 
> where
> import LRSX.Interface.Lexer
> import LRSX.Language.Syntax 
> import LRSX.Language.ContextDefinition
> import Data.List
> import LRSX.Unification.UnificationProblem
> import LRSX.Reduction.Rules
> import Control.Monad.State.Strict
> import qualified Data.Set as Set
> import qualified Data.Map.Strict as Map
> import Data.Char
> import Debug.Trace
> 
> }

> %monad { State ParseState }  { (>>=) } {return } 
> %name parseRuleName RuleName
> %name parseDefinition CtxtDefinition
> %name parseExpr Expr
> %name parse EqConstr
> %name parseEqTwoConstr MatchProblem
> %name parseRules Rules
> %name parseRulesAndCommands RulesAndCommands
> %name parseSRulesAndCommandsInternal SRulesAndCommands
> %name parseConstraints Constraints
> %tokentype { TokenType }
> %token 
> var               { TokenVar  _ _                   }
> string            { TokenString  _ _                 }
> metavar           { TokenMetaVar _ _                }
> '|'               { TokenSymbol   "|" _              }
> '#'               { TokenSymbol   "#" _              }
> '^'               { TokenSymbol   "^" _              }
> '/'               { TokenSymbol   "/" _              }
> ','               { TokenSymbol   "," _              }
> '\\'              { TokenSymbol   "\\" _             }
> ';'               { TokenSymbol   ";" _              }
> '::='             { TokenSymbol "::=" _              }
> '='               { TokenSymbol   "=" _              }
> '/='              { TokenSymbol   "/=" _             }
> '==>'             { TokenSymbol  "==>" _             }
> '.'               { TokenSymbol  "." _               }
> '('               { TokenSymbol  "("  _              }
> '{'               { TokenSymbol  "{"  _              }
> '[.]'             { TokenSymbol  "[.]"  _            }
> '[.'              { TokenSymbol  "[."  _            }
> ')'               { TokenSymbol  ")" _               }
> '<-'              { TokenSymbol  "<-" _               }
> '}'               { TokenSymbol  "}" _               }
> '['               { TokenSymbol  "["  _              }
> ']'               { TokenSymbol  "]"  _              }
> '*'               { TokenSymbol  "*"  _              }
> '+'               { TokenSymbol  "+"  _              }
> ':'               { TokenSymbol  ":"  _              }
> '.l'              { TokenSymbol ".l" _               }
> '.r'              { TokenSymbol ".r" _               }
> '=?='             { TokenSymbol  "=?="  _            }
> 'letrec'          { TokenKeyword "letrec" _          }
> 'flatrec'         { TokenKeyword "flatrec" _          }
> 'in'              { TokenKeyword "in" _              }
> 'overlap'         { TokenKeyword "overlap" _         }
> 'all'             { TokenKeyword "all" _         }
> 'induct'          { TokenKeyword "induct" _         }
> 'unions-from'     { TokenKeyword "unions-from" _    }
> 'atp-path'        { TokenKeyword "atp-path" _    }
> 'and'             { TokenKeyword "and" _             }
> 'HOLE'            { TokenKeyword "HOLE" _            }
> 'var'             { TokenKeyword "var" _            }
> 'where'           { TokenKeyword "where" _           }
> 'except'          { TokenKeyword "except" _          }
> 'restrict'        { TokenKeyword "restrict" _          }
> 'union'           { TokenKeyword "union" _          }
> 'expand'          { TokenKeyword "expand" _          }
> 'ignore'          { TokenKeyword "ignore" _          }
> 'define'          { TokenKeyword "define" _      }
> 'declare'         { TokenKeyword "declare" _      }
> 'prefix'          { TokenKeyword "prefix" _      }
> 'fork'            { TokenKeyword "fork" _      }
> 'ANSWER'          { TokenKeyword "ANSWER" _      }
> 'SR'              { TokenKeyword "SR" _          }
> num               { TokenNumber _ _             }
> %right '.'
> %nonassoc ';'
> %nonassoc 'in' 
> %nonassoc 'letrec'
> %nonassoc 'flatrec'
> %%


> -- =======================================================================================================================
> -- Context-Definitions
> -- =======================================================================================================================
> 



> CtxtDefinition : 'define' CtxtSort '::='  CtxtDefinitionList   {% do 
>                                                                    st <- get
>                                                                    let st1 = (st{psCtxtDefn=(psCtxtDefn st){grammar=(Map.insert $2 $4 (grammar (psCtxtDefn st)))}})
>                                                                    let capturing =  (isCapturingDef $2 (psCtxtDefn st1)) 
>                                                                    let grammnew = Map.fromAscList [(k,setCapturing  $2 g capturing) |   (k,g) <- Map.assocs (grammar (psCtxtDefn st1 ) ) ]
>                                                                    let newfrkTable = setCapturingFrkTable $2 capturing (frkTable (psCtxtDefn st1))
>                                                                    let newpfxTable = setCapturingPfxTable $2 capturing (pfxTable (psCtxtDefn st1))
>                                                                    let st2=st1{psCtxtDefn=(psCtxtDefn st1){grammar=grammnew,frkTable=newfrkTable,pfxTable=newpfxTable}}
>                                                                    (put st2)
>                                                                    
>                                                                }
>                | 'define' CtxtSort {% return ()}
>                | 'declare' 'prefix' metavar metavar '=' Tuple  {%do
>                                                                   st <- get
>                                                                   let pfx1 = takeWhile isAlpha (fromTokenMetaVar $3)
>                                                                   let pfx2 = takeWhile isAlpha (fromTokenMetaVar $4)
>                                                                   if Map.member pfx1 (grammar (psCtxtDefn st)) &&  Map.member pfx2  (grammar (psCtxtDefn st)) then  
>                                                                       do
>                                                                        let k1 = (UserDefinedSort (isCapturingDef pfx1 (psCtxtDefn st)) pfx1)
>                                                                        let k2 = (UserDefinedSort (isCapturingDef pfx2 (psCtxtDefn st)) pfx2)
>                                                                        put (st{psCtxtDefn=(psCtxtDefn st){pfxTable= 
>                                                                              case Map.lookup (k1,k2) (pfxTable (psCtxtDefn st)) of
>                                                                               Nothing -> Map.insert (k1,k2) [$6] (pfxTable (psCtxtDefn st))
>                                                                               Just list -> Map.insert (k1,k2) ($6:list) (pfxTable (psCtxtDefn st))
>                                                                                                           }
>                                                                                })
>                                                                     else if Map.member pfx1 (grammar (psCtxtDefn st)) then error ("context class " ++ show pfx2 ++ " used in declaration, but not defined.")
>                                                                          else error ("context class " ++ show pfx1 ++ " used in declaration, but not defined.")
>                                                         }
>                | 'declare' 'fork' metavar metavar '=' Tuple2   {%do
>                                                                    st <- get
>                                                                    let pfx1 = takeWhile isAlpha (fromTokenMetaVar $3)
>                                                                    let pfx2 = takeWhile isAlpha (fromTokenMetaVar $4)
>                                                                    if Map.member pfx1 (grammar (psCtxtDefn st)) &&  Map.member pfx2  (grammar (psCtxtDefn st)) then  
>                                                                     do
>                                                                      let k1 = (UserDefinedSort (isCapturingDef pfx1 (psCtxtDefn st)) pfx1)
>                                                                      let k2 = (UserDefinedSort (isCapturingDef pfx2 (psCtxtDefn st)) pfx2)
>                                                                      put (st{psCtxtDefn=(psCtxtDefn st){frkTable= 
>                                                                               case Map.lookup (k1,k2) (frkTable (psCtxtDefn st)) of
>                                                                                Nothing -> Map.insert (k1,k2) [$6] (frkTable (psCtxtDefn st))
>                                                                                Just list -> Map.insert (k1,k2) ($6:list) (frkTable (psCtxtDefn st))
>                                                                                                             }
>                                                                               })
>                                                                     else if Map.member pfx1 (grammar (psCtxtDefn st)) then error ("context class " ++ show pfx2 ++ " used in declaration, but not defined.")
>                                                                           else error ("context class " ++ show pfx1 ++ " used in declaration, but not defined.")
>               
>                                                         }
>  
> Tuple  : '(' metavar ',' metavar ')' {% do
>                                          state <- get
>                                          let pfx1 = takeWhile isAlpha (fromTokenMetaVar $2)
>                                          let pfx2 = takeWhile isAlpha (fromTokenMetaVar $4)
>                                          if Map.member pfx1 (grammar (psCtxtDefn state)) &&  Map.member pfx2  (grammar (psCtxtDefn state)) then 
>                                             return (UserDefinedSort (isCapturingDef pfx1 (psCtxtDefn state)) pfx1, UserDefinedSort (isCapturingDef pfx2 (psCtxtDefn state)) pfx2)
>                                           else if Map.member pfx1 (grammar (psCtxtDefn state)) then error ("context class " ++ show pfx2 ++ " used in declaration, but not defined.")
>                                                 else error ("context class " ++ show pfx1 ++ " used in declaration, but not defined.")
>                                      }
> Tuple2  : '(' metavar ',' metavar ',' metavar ','  Expr ')' {% do
>                                          state <- get
>                                          let pfx1 = takeWhile isAlpha (fromTokenMetaVar $2)
>                                          let pfx2 = takeWhile isAlpha (fromTokenMetaVar $4)
>                                          let pfx3 = takeWhile isAlpha (fromTokenMetaVar $6)
>                                          case () of 
>                                           _ | not (Map.member pfx1 (grammar (psCtxtDefn state))) ->  error ("context class " ++ show pfx1 ++ " used in declaration, but not defined.")
>                                             | not (Map.member pfx2 (grammar (psCtxtDefn state))) ->  error ("context class " ++ show pfx2 ++ " used in declaration, but not defined.")
>                                             | not (Map.member pfx3 (grammar (psCtxtDefn state))) ->  error ("context class " ++ show pfx3 ++ " used in declaration, but not defined.")
>                                             | otherwise ->   return (UserDefinedSort (isCapturingDef pfx1 (psCtxtDefn state)) pfx1
>                                                                     ,UserDefinedSort (isCapturingDef pfx2 (psCtxtDefn state)) pfx2
>                                                                     ,UserDefinedSort (isCapturingDef pfx3 (psCtxtDefn state)) pfx3, $8) }

  
  
> -- Note that we have to parse the CtxtSort by an extra non-terminal
> -- to first execute the action, which adds the name of the new sort
> -- to the known context sorts
>
> CtxtSort : metavar  {% do 
>                                                           st <- get
>                                                           case Map.lookup (fromTokenMetaVar  $1) (grammar $ psCtxtDefn st) of 
>                                                             Just [] -> do
>                                                              put (st{psCtxtDefn=(psCtxtDefn st){grammar=(Map.insert (fromTokenMetaVar $1) [] (grammar (psCtxtDefn st)))}})
>                                                              return (fromTokenMetaVar $1)
>                                                             Just _ -> error ("Ctxt-Symbol" ++ show $1 ++ " already defined")
>                                                             Nothing -> do
>                                                              put (st{psCtxtDefn=(psCtxtDefn st){grammar=(Map.insert (fromTokenMetaVar $1) [] (grammar (psCtxtDefn st)))}})
>                                                              return (fromTokenMetaVar $1)
>                                                       }

> CtxtDefinitionList : CExpr                         { [$1] }
>                    | CtxtDefinitionList '|' CExpr  { $3:$1 }

> CExpr:               Expr                           { let e = $1 
>                                                       in if hasHole e 
>                                                            then Right (e,[],[],[]) 
>                                                            else error ("Expression " ++ show $1 ++ " instead of context ") 
>                                                     }
>                    | Expr '/' Expr                  { let 
>                                                           e1 = $1 
>                                                           e2 = $3
>                                                       in 
>                                                          if hasHole e1 && hasHole e2  
>                                                               then Left ((e1,[],[],[]),(e2,[],[],[])) 
>                                                               else error ("Expressions " ++ show (e1,e2) ++ " instead of context ") }
>                    | Expr  'where' Constraints      { let e = $1 
>                                                       in 
>                                                         if hasHole e  
>                                                          then Right (e,ctxtcon $3,envcon $3,capcon $3) 
>                                                          else error ("Expression " ++ show $1 ++ " instead of context ") }
>                    | Expr  'where' Constraints '/' Expr  'where' Constraints 
>                                                     { let 
>                                                           e1 = $1 
>                                                           e2 = $5
>                                                       in   
>                                                         if hasHole e1 && hasHole e2
>                                                            then Left ((e1,ctxtcon $3,envcon $3,capcon $3),(e2,ctxtcon $7,envcon $7,capcon $7))
>                                                            else error ("Expressions " ++ show (e1,e2) ++ " instead of context ") }




> -- =======================================================================================================================

> SRulesAndCommands : SRules                               {% return ()} 


> RulesAndCommands ::  {()}
> RulesAndCommands : Rules                               {% return () } 
> Dir : '.l' { Left }
>     | '.r' { Right }
> Option : 'ignore' VLIST                                { ($2,[]) }
>        | 'restrict' VILIST                             { ([],$2) }
>        | 'ignore' VLIST 'restrict' VILIST              { ($2,$4) }
>

> OverlapCommand : 'overlap' RuleNameInCommand Dir 'and' RuleNameInCommand Dir {Single ($3 $2) ($6 $5) [] []}
>          | 'overlap' RuleNameInCommand Dir 'and' RuleNameInCommand Dir Option  {let (ignored,restricted) = $7 in Single ($3 $2) ($6 $5) ignored restricted}
>          | 'overlap' RuleNameInCommand Dir 'all'          {All ($3 $2) [] [] []}
>          | 'overlap' RuleNameInCommand Dir 'all' Option   {let (ignored,restricted) = $5 in All ($3 $2) [] ignored restricted}
>          | 'overlap' RuleNameInCommand Dir 'all' 'except' VLIST {All ($3 $2) $6 [] []}
>          | 'overlap' RuleNameInCommand Dir 'all' 'except' VLIST Option {let (ignored,restricted) = $7 in All ($3 $2) $6 ignored restricted}

> Command  : OverlapCommand                                 {% do 
>                                                               state <- get 
>                                                               put state{psOverlapCommands = (psOverlapCommands state) ++ [$1 Nothing]}
>                                                           }
>          | string '<-' OverlapCommand                     {% do 
>                                                               state <- get 
>                                                               put state{psOverlapCommands = (psOverlapCommands state) ++ [$3 (Just $ fromTokenString $1)]}
>                                                           }
>          | 'union'  RuleNameInCommand '=' '{' VarList '}' {% do 
>                                                                 state <- get
>                                                                 put state{psUnionCommands = (psUnionCommands state) ++ [UNION $2 $5]}
>                                                           }
>          | 'expand' RuleNameInCommand '=' '{' VarList '}' {% do 
>                                                                 state <- get
>                                                                 put state{psExpandCommands = (psExpandCommands state) ++ [EXPAND $2 $5 Nothing]}
>                                                           }
>          | 'induct' string 'atp-path' '=' string 'unions-from' '=' string {% do
>                                                                               state <- get
>                                                                               put state {psInductCommands = (psInductCommands state) ++ [(fromTokenString $2,fromTokenString $5,fromTokenString $8)]}
>                                                                           }



> VILIST : RuleNameInCommand ':' num              {[($1,fromTokenNumber $3)]}
>       | '{' VarIList '}'   {$2}

> VarIList : RuleNameInCommand ':' num              { [($1,fromTokenNumber $3)]  }
>         | RuleNameInCommand ':' num ',' VarIList  { ($1,fromTokenNumber $3):$5 }

> VLIST : RuleNameInCommand               {[$1]}
>       | '{' VarList '}'   {$2}

>  VarList : RuleNameInCommand             {[$1]}
>         | RuleNameInCommand ',' VarList {$1:$3}

> SRules :: {()}
> SRules : Rules   {% return () }

Rules = List of 4-Tupel (rule name, rule, list of nonempty contexts, list of nonempty enviroments)

> Rules :: {()}
> Rules : RuleOrAnswer                 {% return ()}
>       | RuleOrAnswer  Rules          {% return ()}
> RuleNameInCommand :: {RuleName}
> RuleNameInCommand : '(' var                         ')'       {(False,fromTokenVar $2,1,False)}
>                   | '(''SR' ',' var                 ')'       {(True,fromTokenVar $4, 1,False)}
>                   | '(' var ',' num                 ')'       {(False,fromTokenVar $2, fromTokenNumber $4,False)}
>                   | '(''SR' ',' var ',' num         ')'       {(True,fromTokenVar $4,fromTokenNumber $6,False)} 
>                   | '(''ANSWER' ',' num             ')'       {(False,"ANSWER",fromTokenNumber $4,False)} 
>                   | '(' var ',' '+'                 ')'       {(False,fromTokenVar $2,1,True)}
>                   | '(''SR' ',' var ',' '+' ',' num ')'       {(True,fromTokenVar $4,fromTokenNumber $8,True)}
>                   | '(' var ',' '+' ',' num         ')'       {(False,fromTokenVar $2,fromTokenNumber $6,True)}
>                   
> RuleName : '{' var '}'                                        { (0,False,fromTokenVar $2,1,False)  }
>          | '{' 'SR' ',' var '}'                               { (0,True,fromTokenVar $4,1,False) }
>          | '{' num ',' 'SR' ',' var '}'                       { (fromTokenNumber $2,True,fromTokenVar $6,1,False) }
>          | '{' var ','  num '}'                               { (0,False,fromTokenVar $2,fromTokenNumber $4,False)  }
>          | '{' 'SR' ',' var ',' num '}'                       { (0,True,fromTokenVar $4,fromTokenNumber $6,False) }
>          | '{' num ',' 'SR' ',' var ',' num '}'               { (fromTokenNumber $2,True,fromTokenVar $6,fromTokenNumber $8,False) }
>          | '{' var ',' '+' '}'                                { (0,False,fromTokenVar $2,1,True)  }
>          | '{' 'SR' ',' var ',' '+' '}'                       { (0,True,fromTokenVar $4,1,True) }
>          | '{' num ',' 'SR' ',' var ',' '+'  '}'              { (fromTokenNumber $2,True,fromTokenVar $6,1,True) }
>          | '{' var ',' '+' ','  num '}'                       { (0,False,fromTokenVar $2,fromTokenNumber $6,True)  }
>          | '{' 'SR' ',' var ',' '+' ',' num '}'               { (0,True,fromTokenVar $4,fromTokenNumber $8,True) }
>          | '{' num ',' 'SR' ',' var ',' '+' ',' num '}'       { (fromTokenNumber $2,True,fromTokenVar $6,fromTokenNumber $10,True) }

> RuleOrAnswer : Rule                                           {% return () }
>              | 'ANSWER' '{' var ',' num '}' Answer            {% do 
>                                                                   let r = $7
>                                                                   let ansRule = (r{ruleName=(False,fromTokenVar $3,fromTokenNumber $5,False)}) 
>                                                                   state <- get
>                                                                   put state{psAnswerRules=ansRule:(psAnswerRules state)}
>                                                               }
>                                                                    
>              | 'ANSWER' '{' var '}' Answer                    {% do 
>                                                                   let r = $5 
>                                                                   state <- get
>                                                                   let c = psCounter state
>                                                                   let ansRule = (r{ruleName=(False,fromTokenVar $3,c,False)})  
>                                                                   put state{psAnswerRules=ansRule:(psAnswerRules state),psCounter=c+1}
>                                                               }
>              | 'ANSWER' Answer                                {% do 
>                                                                   let (a,b,_,d) = ruleName $2
>                                                                   state <- get
>                                                                   let c = psCounter state
>                                                                   let ansRule = ($2{ruleName=(a,b,c,d)})  
>                                                                   put state{psAnswerRules=ansRule:(psAnswerRules state),psCounter=c+1}
>                                                               }
>             | Command                                         {% return ()}
>             | CtxtDefinition {% do return ()}  
>
> Answer : Expr                    { let e = $1 
>                                    in if isExpr e then
>                                         ConstrainedRule {
>                                              ruleName = (False,"ANSWER",0,False)
>                                             ,rule = e :==>: (Fn "ANSWER" [])
>                                             ,nonEmptyCtxts = []
>                                             ,nonEmptyEnvs  = []
>                                             ,nonCapCon     = []
>                                             ,priority = 0
>                                             ,standardReduction = False
>                                             ,answerRule = True
>                                         }
>                                        else error ("expression " ++ show e ++ " is not an expression, but a context")
>                                  }
>        | Expr 'where' Constraints { let e = $1 
>                                     in if isExpr e then
>                                                             (ConstrainedRule {
>                                                                ruleName = (False,"ANSWER",0,False)
>                                                               ,rule = e :==>: (Fn "ANSWER" [])
>                                                               ,nonEmptyCtxts = ctxtcon $3
>                                                               ,nonEmptyEnvs  = envcon $3
>                                                               ,nonCapCon     = capcon $3  
>                                                               ,priority = 0
>                                                               ,standardReduction = False
>                                                               ,answerRule = True
>                                                               }
>                                                              )
>                                        else error ("expression " ++ show e ++ " is not an expression, but a context")
>                                                       }
> Rule : RuleName Expr '==>' Expr {% do
>                                     let e1 = $2
>                                     let e2 = $4
>                                     case () of
>                                         _ | isExpr e1 && isExpr e2 -> do
>                                                                        let isSR = let (prio,sr,name,number,trans) = $1 in sr
>                                                                        let newRule = ConstrainedRule {
>                                                                                                ruleName = let (prio,sr,name,number,trans) = $1 in (sr,name,number,trans) 
>                                                                                               ,rule = e1 :==>: e2
>                                                                                               ,nonEmptyCtxts = []
>                                                                                               ,nonEmptyEnvs  = []
>                                                                                               ,nonCapCon     = []
>                                                                                               ,priority = let (prio,sr,name,number,trans) = $1 in prio
>                                                                                               ,standardReduction = let (prio,sr,name,number,trans) = $1 in sr
>                                                                                               ,answerRule = False
>                                                                                      }
>                                                                        state <- get
>                                                                        put state{psStandardReductions = if isSR then ((psStandardReductions state)++[newRule]) else (psStandardReductions state)
>                                                                                 ,psTransformations    = if isSR then (psTransformations state) else ((psTransformations state)++[newRule])
>                                                                                 }  
>                                           | not (isExpr e2) ->  error ("expression " ++ show e2 ++ " is not an expression, but a context")
>                                           | otherwise       ->  error ("expression " ++ show e1 ++ " is not an expression, but a context") 
>                                 }
>
>      | RuleName  Expr '==>' Expr 'where' Constraints  { % do
>                                                            let e1 = $2
>                                                            let e2 = $4
>                                                            case () of 
>                                                              _ | isExpr e1 && isExpr e2 ->do
>                                                                        let isSR = let (prio,sr,name,number,trans) = $1 in sr
>                                                                        let newRule = ConstrainedRule {
>                                                                                              ruleName = let (prio,sr,name,number,trans) = $1 in (sr,name,number,trans)
>                                                                                             ,rule = e1 :==>: e2
>                                                                                             ,nonEmptyCtxts = ctxtcon $6
>                                                                                             ,nonEmptyEnvs  = envcon $6
>                                                                                             ,nonCapCon     = capcon $6
>                                                                                             ,priority = let (prio,sr,name,number,trans) = $1 in prio
>                                                                                             ,standardReduction = let (prio,sr,name,number,trans) = $1 in sr
>                                                                                             ,answerRule = False
>                                                                                       }
>                                                                        state <- get
>                                                                        put state{psStandardReductions = if isSR then ((psStandardReductions state)++[newRule]) else (psStandardReductions state)
>                                                                                 ,psTransformations    = if isSR then (psTransformations state) else (psTransformations state)++[newRule]
>                                                                                 }  
>                                                                | not (isExpr e2)  -> error ("expression " ++ show e2 ++ " is not an expression, but a context")
>                                                                | otherwise        -> error ("expression " ++ show e1 ++ " is not an expression, but a context")
>                                                       }


==========================================================================================
Parsing Equations with Constraints, to use for pure unification
==========================================================================================
==
==

EqConstraint = List of equations, list of nonempty contexts, list of nonempty environments

> EqConstr :: {(EquationsWithConstraints)}
> EqConstr : CtxtDefinitions  '#' EqConstr {$3}
>          | Equations                     { let eqs = $1 in (eqs,[],[],[])}
>          | Equations 'where' Constraints { let eqs =$1 in ((eqs,ctxtcon $3,envcon $3, capcon $3))}



> Equations :: {([Equation Expression])}
> Equations : Equation {[$1]}
>           | Equation ',' Equations {$1:$3}
> Equation : Expr '=?=' Expr    { let 
>                                   e1 = $1
>                                   e2 = $3
>                                 in
>                                   if isExpr e1 then 
>                                    if isExpr e2   then e1 :=?=: e2
>                                                   else error ("expression " ++ show e2 ++ " is not an expression, but a context" )
>                                    else error ("expression " ++ show e1 ++ " is not an expression, but a context" )
>                               }

==
==
==========================================================================================


==========================================================================================
Parsing Equations with Constraints, to use for matching (experimental)
==========================================================================================
==
==

> CtxtDefinitions : CtxtDefinition {% return ()}
>                 | CtxtDefinition CtxtDefinitions {% return ()}
> MatchProblem : CtxtDefinitions '#'  EqTwoConstr  {$3}
>              | EqTwoConstr {$1}
> EqTwoConstr :: {(EquationsWithTwoConstraints)}
> EqTwoConstr : Equations                                  { let  eqs=$1 in ((eqs,[],[],[],[],[],[]))}
>          | Equations 'where' '|' Constraints             { let eqs=$1 in ((eqs,[],[],[],ctxtcon $4,envcon $4,capcon $4)) }
>          | Equations 'where' Constraints '|'             { let eqs=$1 in ((eqs,ctxtcon $3,envcon $3, capcon $3,[],[],[]))  }
>          | Equations 'where' Constraints '|' Constraints { let eqs=$1 in ((eqs,ctxtcon $3,envcon $3, capcon $3,ctxtcon $5,envcon $5,capcon $5))}



==
==
==========================================================================================


==========================================================================================
Parsing Expressions
together with strictness information of function symbol positions
==========================================================================================
==
==



== Productions for Expr



> Expr :: { Expression }
> Expr : '\\' LamVar '.' Expr           { % do
>                                             let lamv = $2
>                                             state <- get
>                                             case Map.lookup "\\" (psFunInfo state) of 
>                                               Nothing -> put state{psFunInfo =  Map.insert "\\" (1,[1]) (psFunInfo state)}
>                                               Just (ari,bari) -> if (ari,bari) == (1,[1]) then put state else (error (printMark [$1] ++ " " ++ "different function information for \\: " ++ show (ari,bari) ++ "/= (1,[],[1])"))
>                                             return (Fn "\\" [Binder [lamv] $4])
>                                       }
>
>      | 'letrec' '{' Env '}' 'in' Expr {  (Letrec (applyEnvItems $3 empty) $6) }
> 
>      | 'letrec' Env 'in' Expr         {  (Letrec (applyEnvItems $2 empty) $4) }
>      | 'flatrec' '{' Env '}' 'in' Expr {  (FlatLetrec (applyEnvItems $3 empty) $6) }
> 
>      | 'flatrec' Env 'in' Expr         {  (FlatLetrec (applyEnvItems $2 empty) $4) }
>      | SExpr                          { $1 }                   

== Productions for SExpr

> SExpr :: {(Expression)}
> SExpr : var LExpr                     {% do 
>                                           let argsWithBindArities = $2 
>                                           let rev = reverse argsWithBindArities
>                                           let args = map fst rev
>                                           let bindarities = map snd rev
>                                           let finfo :: (Arity,[BindArity]); finfo = (length args,bindarities) 
>                                          
>                                           state <- get 
>                                           case Map.lookup (fromTokenVar $1) (psFunInfo state) of 
>                                               Nothing -> put state{psFunInfo =  Map.insert (fromTokenVar $1) finfo (psFunInfo state)}
>                                               Just (ari,bari) -> if (ari,bari) == (length args,bindarities) 
>                                                                         then put state 
>                                                                         else (error ( printMark [$1] ++ " " ++ "different function information for" ++ showToken $1 ++":" ++ show (ari,bari) ++ "/= " ++ (show (length args,bindarities))))
>                                           return (Fn (fromTokenVar $1) args) 
>                                       }

>       | var '[' ']'                   {% do
>                                             state <- get
>                                             let finfo = (0,[])
>                                             case Map.lookup (fromTokenVar $1)  (psFunInfo state) of 
>                                               Nothing -> put state{psFunInfo =  Map.insert (fromTokenVar $1)  finfo (psFunInfo state)}
>                                               Just (ari,bari) -> if (ari,bari) == finfo 
>                                                                         then put state 
>                                                                         else (error $ printMark [$1] ++ " " ++ "different function information for" ++ showToken $1 ++":" ++ show (ari,bari) ++ "/= " ++ "(0,[])")
>                                             return (Fn (fromTokenVar $1) [])
>                                       }
>       | var '[' SPos ']' LExpr        {% do
>                                             let argsWithBindArities = $5 
>                                             let rev = reverse argsWithBindArities
>                                             let args = map fst rev
>                                             let bindarities = map snd rev
>                                             state <- get
>                                             let finfo = (length args,bindarities)
>                                             case Map.lookup (fromTokenVar $1)  (psFunInfo state) of 
>                                               Nothing -> put state{psFunInfo =  Map.insert (fromTokenVar $1)  finfo (psFunInfo state)}
>                                               Just (ari,bari) -> if (ari,bari) == finfo 
>                                                                         then put state 
>                                                                         else (error $ printMark [$1] ++ " " ++  "different function information for" ++ showToken $1 ++":" ++ show (ari,bari) ++ "/= " ++ (show (length args,$3,bindarities)))
>                                             return (Fn (fromTokenVar $1) args)
>                                       }
>       | var '['  ']' LExpr            {% do
>                                             let argsWithBindArities = $4
>                                             let rev = reverse argsWithBindArities
>                                             let args = map fst rev
>                                             let bindarities = map snd rev
>                                             state <- get
>                                             let finfo = (length args,bindarities)
>                                             case Map.lookup (fromTokenVar $1)  (psFunInfo state) of 
>                                               Nothing -> put state{psFunInfo =  Map.insert (fromTokenVar $1)  finfo (psFunInfo state)}
>                                               Just (ari,bari) -> if (ari,bari) == finfo 
>                                                                         then put state 
>                                                                         else (error $ printMark [$1] ++ " " ++  "different function information for" ++ showToken $1 ++":" ++ show (ari,bari) ++ "/= " ++ (show (length args,bindarities)))
>                                             return (Fn (fromTokenVar $1) args)
>                                       }
>       | Atom                          { $1 }
 



> LExpr :: {[(ArgPos,BindArity)]}
> LExpr : Binder                           {  [$1] }
>       | LExpr Binder                     {  (($2):$1) }

== Productions for Binder

> Binder :: {(ArgPos,BindArity)}
> Binder : Atom                            { (Binder [] $1,0) }
>        | LamVar '.' Binder               { let (Binder xs e,i) = $3 in (Binder ($1:xs) e,i+1) }
>        | '*' LamVar                      { (VarPos $2,-1)}
>        | '(' '*' LamVar ')'              { (VarPos $3,-1)}
>        | '(' LamVar '.' Binder ')'       { let (Binder xs e,i) = $4 in (Binder ($2:xs) e,i+1) }

== Productions for Atom

> Atom :: {Expression}
> Atom:  'var' LamVar                       { (VarL $2)}
>       | var                                      {% do
>                                                     state <- get
>                                                     let finfo = (0,[])
>                                                     case Map.lookup (fromTokenVar $1) (psFunInfo state) of 
>                                                        Nothing -> put state{psFunInfo =  Map.insert (fromTokenVar $1)  finfo (psFunInfo state)}
>                                                        Just (ari,bari) -> if (ari,bari) == finfo 
>                                                                            then put state 
>                                                                            else (error $ printMark [$1] ++ " " ++  "different function information for" ++ showToken $1 ++":" ++ show (ari,bari) ++ "/= " ++ "(0,[])")
>                                                     return (Fn (fromTokenVar $1) [])
>                                                   }

>      | metavar '[' Expr ']'              {% do
>                                               state <- get
>                                               let pfx = (takeWhile isAlpha (fromTokenMetaVar $1))
>                                               if Map.member pfx (grammar (psCtxtDefn state))
>                                                then return (let e = $3 in (MetaCtxtSubst emptySubstitution (UserDefinedSort (isCapturingDef pfx (psCtxtDefn state)) pfx) (fromTokenMetaVar $1) e))
>                                                else error (printMark [$1] ++ " " ++ "Ctxt " ++  pfx ++ showToken $1 ++ "used but not defined")
>                                          }
>      | metavar                           {% do 
>                                              state <- get
>                                              let pfx = (takeWhile isAlpha (fromTokenMetaVar $1))
>                                              if Map.member pfx (grammar (psCtxtDefn state))
>                                               then return (MetaCtxt emptySubstitution (UserDefinedSort (isCapturingDef pfx (psCtxtDefn state)) pfx) (fromTokenMetaVar $1))
>                                               else return (MetaExpression emptySubstitution (fromTokenMetaVar $1))}
>      | '(' Expr ')'                      { $2 }
>      | 'HOLE'                            { (Hole emptySubstitution) }
>      | 'HOLE' num                        { (MultiHole (fromTokenNumber $2)) }
>      | '[.]'                             { (Hole emptySubstitution) }
>      | '[.' num ']'                      { (MultiHole (fromTokenNumber $2)) }


== Productions for LamVar 

> LamVar  : metavar                         { MetaVarLift emptySubstitution (fromTokenMetaVar $1) } 
>         | var                             { VarLift emptySubstitution  (fromTokenVar $1) }

== Productions for LetVar 


> Env : EnvItem                                 { [$1] }
>     | EnvItem ';' Env                         { $1:$3 }

> EnvItem : metavar                             { (A $ fromTokenMetaVar $1) }
>         | metavar '^' metavar '[' LamVar ',' Expr ']'     { (B (MetaChain emptySubstitution  (UserDefinedChain (fromTokenMetaVar $3)) (fromTokenMetaVar $1) $5 $7)) }
>         | metavar '|' LamVar ',' Expr '|'     { let e = $5 in (B (MetaChain emptySubstitution  VarChain (fromTokenMetaVar $1) $3 e)) }
>         | LamVar '=' Expr                     { let e = $3 in (C ($1 := e)) }

== Productions for SPos

> SPos :: {[Int]}
> SPos : num            {[fromTokenNumber $1]}
>      | num ',' SPos   {(fromTokenNumber $1):$3}


==
==
==========================================================================================


==========================================================================================
Parsing Constraints
==========================================================================================
==
==

> Constraints : Constraint {[$1]}
>             | Constraint ',' Constraints {$1:$3}
> Constraint :  metavar  '/=' '{' '}'       { (EnvConstraint  (fromTokenMetaVar $1))
>                                           }
>            |  metavar  '/=' '[' ']'       {%  do
>                                                state <- get
>                                                let pfx = (takeWhile isAlpha (fromTokenMetaVar $1))
>                                                if Map.member pfx (grammar (psCtxtDefn state)) then 
>                                                    return (CtxtConstraint ((isCapturingDef (takeWhile isAlpha (fromTokenMetaVar $1)) (psCtxtDefn state)),(fromTokenMetaVar $1)))
>                                                 else error ("context class: " ++ show pfx ++ "used in constraint, but not defined.")
>                                           }
>            |  metavar  '/=' '[.]'        {%  do
>                                                state <- get
>                                                let pfx = (takeWhile isAlpha (fromTokenMetaVar $1))
>                                                if Map.member pfx (grammar (psCtxtDefn state)) then 
>                                                    return (CtxtConstraint ((isCapturingDef (takeWhile isAlpha (fromTokenMetaVar $1)) (psCtxtDefn state)),(fromTokenMetaVar $1)))
>                                                 else error ("context class: " ++ show pfx ++ "used in constraint, but not defined.")
>                                          }
>            |  metavar  '/=' 'HOLE'        {%  do
>                                                state <- get
>                                                let pfx = (takeWhile isAlpha (fromTokenMetaVar $1))
>                                                if Map.member pfx (grammar (psCtxtDefn state)) then 
>                                                    return (CtxtConstraint ((isCapturingDef (takeWhile isAlpha (fromTokenMetaVar $1)) (psCtxtDefn state)),(fromTokenMetaVar $1)))
>                                                 else error ("context class: " ++ show pfx ++ "used in constraint, but not defined.")
>                                           }

>            |  '(' Expr ',' Expr ')'   {let 
>                                            e1 = $2 
>                                            e2 = $4 
>                                        in CapConstraint  (e1,e2) }
>            |  '[' Env ',' Expr ']'    {let 
>                                            e1 = $2 
>                                            e2 = $4 
>                                        in CapConstraintEnv  (e1,e2) }

==
==
==========================================================================================


> {

help functions, for computing the enviroment and context constraints

> data Constraint a b c d = EnvConstraint a | CtxtConstraint b | CapConstraint c | CapConstraintEnv d
> 
> envcon xs = [Env emptySubstitution x | (EnvConstraint x) <- xs]
> ctxtcon xs = [MetaCtxt emptySubstitution (UserDefinedSort b (takeWhile isAlpha x)) x | (CtxtConstraint (b,x)) <- xs]
> capcon xs = [(Left a,b) | (CapConstraint (a,b)) <- xs]
>            ++ [(Right $ applyEnvItems a empty,b) | CapConstraintEnv (a,b) <- xs]
>

help functions, for computing the parts of an enviroment

> data EnvItem a b c = A a | B b | C c
> applyEnvItems xs env = 
>   env{metaEnvVars = map (Env emptySubstitution) (map fromA $ filter isA xs),
>       metaChains  = map fromB $ filter isB xs,
>       bindings    = map fromC $ filter isC xs}
>  where
>   isA (A _) = True
>   isA _ = False
>   isB (B _) = True
>   isB _ = False
>   isC (C _) = True
>   isC _ = False
>   fromA (A x) = x
>   fromB (B x) = x
>   fromC (C x) = x

the error function

> happyError :: [TokenType] -> a
> happyError []     = error $ "parse error: unexpected end of file "
> happyError (x:xs) = error $ "parse error before: "
>                           ++ (take 200 $ concat ( intersperse " " ( map showToken xs)))  ++ (printMark xs)




> parser str =
>  let
>   (parseRes,st) = runState (parse $ lexInput str) initState
>   funinfo = toFunInfo (psFunInfo st)
>  in case (check (parseResToProblem parseRes)) of 
>        Just err -> error err
>        Nothing -> (parseRes,funinfo,psCtxtDefn st)

> parseResToProblem (eqs,ctxtcon,envcon,ncc) =
>    Problem {                         solution       = []
>                                     ,gamma          = gam
>                                     ,delta          = Delta {delta1=ctxtcon,delta2=envcon,delta3=ncc,delta4=emptyGamma}
>                                     ,fresh          = []
>                                     ,functions      = []
>                                     ,inputGamma     = gam
>                                     ,inputDelta     = Delta {delta1=ctxtcon,delta2=envcon,delta3=ncc,delta4=emptyGamma}
>                                     ,nabla   = emptyDelta
>                                 }
>   where gam =  Gamma { varLiftEq = []
>                      , expressionEq = eqs
>                      , environmentEq = []
>                      , bindingEq = []
>                      } 




> compactFunInfo :: FunInfo -> FunInfo
> compactFunInfo finfo =
>  let 
>       sorted = sortBy (\(n,arity,bari) (n1,arity1,bari1) -> compare n n1) finfo
>       grouped = groupBy (\(n,arity,bari) (n1,arity1,bari1) -> n == n1) sorted
>       joinGroup group@((n,arity,bari):_) = 
>                    let 
>                       arities = [a | (_,a,_) <- group]
>                       bindarities  = [b | (_,_,b) <- group]
>                       lbindarities = [b | b <- bindarities, b /= []]
>                       theArity = if (length (nub arities) /= 1) then
>                                       error $ "function symbol '" ++ n ++ "' occurs with different arities"
>                                   else arity
>                       theBindArities = if null lbindarities then [] else
>                                         if (length (nub lbindarities)) == 1 then (head lbindarities) else
>                                       error $ "function symbol '" ++ n ++ "' occurs with different binder arities"
>                          
>                    in (n,theArity,theBindArities)
>  in map joinGroup grouped
> isUNION (UNION _ _) = True
> isUNION _ = False
> isExpand (EXPAND _ _ _) = True
> isExpand _ = False
>
>
> data ParseState = ParseState {psCtxtDefn           :: CtxtDef
>                              ,psFunInfo            :: Map.Map Name (Arity,[BindArity])
>                              ,psStandardReductions :: [ConstrainedRule]
>                              ,psAnswerRules        :: [ConstrainedRule]
>                              ,psTransformations    :: [ConstrainedRule]
>                              ,psCounter            :: Int
>                              ,psOverlapCommands    :: [Command]
>                              ,psUnionCommands      :: [Command]
>                              ,psExpandCommands     :: [Command]
>                              ,psInductCommands     :: [(String,String,String)]
>                              }
>  deriving(Show)
> initState = ParseState{psCtxtDefn             = initCtxtDef
>                       ,psFunInfo              = Map.empty
>                       ,psStandardReductions   = []
>                       ,psAnswerRules          = []
>                       ,psTransformations      = []
>                       ,psCounter              = 0
>                       ,psOverlapCommands      = []
>                       ,psUnionCommands        = []
>                       ,psExpandCommands       = []
>                       ,psInductCommands       = []
> }
> runP p =  evalState p initState 
> runPST st p =  evalState p st
> runStatePST st p =  runState p st
>
> parserMatching str = parseMatchProblem $ lexInput str
> parseMatchProblem :: [TokenType] -> (EquationsWithTwoConstraints,FunInfo,CtxtDef)
> parseMatchProblem tk = let (a,s) = runState (parseEqTwoConstr tk) initState
>                        in (a,toFunInfo $ psFunInfo s,psCtxtDefn s)
> toFunInfo mp = [(a,b,d) |  (a,(b,d)) <-  Map.assocs mp]

 

> parseSRulesAndCommands tk = runState (parseSRulesAndCommandsInternal tk) initState
> onlyConstraints state tk  = let (as,st) = runState (parseConstraints tk) state
>                             in  (ctxtcon as, envcon as, capcon as, st)



> }


